import yaml
from typing import Optional


EMAIL_TEMPLATE = "<your@mail.com>"
FULL_NAME_TEMPLATE = "<Your Full Name for gitea>"
PASSWORD_TEMPLATE = "<Your password for gitea>"
USERNAME_TEMPLATE = "<Your username for gitea>"

class color:
   PURPLE = '\033[95m'
   CYAN = '\033[96m'
   DARKCYAN = '\033[36m'
   BLUE = '\033[94m'
   GREEN = '\033[92m'
   YELLOW = '\033[93m'
   RED = '\033[91m'
   BOLD = '\033[1m'
   UNDERLINE = '\033[4m'
   END = '\033[0m'


def read_config(conf_path: str):
    with open(conf_path) as f:
        config = yaml.safe_load(f)
    return config


def add_to_config(conf_path: str, data, key):
    with open(conf_path) as file:
        try:
            list_doc = yaml.safe_load(file)
        except yaml.YAMLError as exc:
            print(exc)
    list_doc[key] = data
    with open(conf_path, "w") as f:
        yaml.dump(list_doc, f)


def create_template_user_config(conf_path: str,
                                email: Optional[str] = None,
                                full_name: Optional[str] = None,
                                password: Optional[str] = None,
                                username: Optional[str] = None,
                                ):
    data = {
        "email": EMAIL_TEMPLATE if email is None else email,
        "full_name": FULL_NAME_TEMPLATE if full_name is None else full_name,
        "password": PASSWORD_TEMPLATE if password is None else password,
        "username": USERNAME_TEMPLATE if username is None else username,
    }
    with open(conf_path, "w") as f:
        yaml.dump(data, f)


from __future__ import annotations

import argparse
import dataclasses
import io
import json
import os
import platform
import re
import shutil
import site
import subprocess
import sys
import urllib.request
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Sequence

if sys.version_info < (3, 12):
    sys.exit("Python 3.12 or above is required to install trusted-ai.")

_plat = platform.system()
MACOS = _plat == "Darwin"
JSON_URL = "https://pypi.org/pypi/trusted-ai/json"

FOREGROUND_COLORS = {
    "black": 30,
    "red": 31,
    "green": 32,
    "yellow": 33,
    "blue": 34,
    "magenta": 35,
    "cyan": 36,
    "white": 37,
}


def _call_subprocess(args: list[str]) -> int:
    try:
        return subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, check=True).returncode
    except subprocess.CalledProcessError as e:
        print(f"An error occurred when executing {args}:", file=sys.stderr)
        print(e.output.decode("utf-8"), file=sys.stderr)
        sys.exit(e.returncode)


def _echo(text: str) -> None:
    sys.stdout.write(text + "\n")


def _add_to_path(target: Path) -> None:
    value = os.path.normcase(target)

    paths = [os.path.normcase(p) for p in os.getenv("PATH", "").split(os.pathsep)]
    if value in paths:
        return
    _echo(
        "Post-install: Please add {} to PATH by executing:\n    {}".format(
            colored("green", value),
            colored("cyan", f"export PATH={value}:$PATH"),
        )
    )


def support_ansi() -> bool:
    if not hasattr(sys.stdout, "fileno"):
        return False

    try:
        return os.isatty(sys.stdout.fileno())
    except io.UnsupportedOperation:
        return False


def colored(color: str, text: str, bold: bool = False) -> str:
    if not support_ansi():
        return text
    codes = [FOREGROUND_COLORS[color]]
    if bold:
        codes.append(1)

    return "\x1b[{}m{}\x1b[0m".format(";".join(map(str, codes)), text)


@dataclasses.dataclass
class Installer:
    location: str | None = None
    version: str | None = None
    prerelease: bool = False
    additional_deps: Sequence[str] = ()
    skip_add_to_path: bool = False
    output_path: str | None = None

    def __post_init__(self):
        self._path = self._decide_path()
        self._path.mkdir(parents=True, exist_ok=True)
        if self.version is None:
            self.version = self._get_latest_version()

    def _get_latest_version(self) -> str:
        resp = urllib.request.urlopen(JSON_URL)
        metadata = json.load(resp)

        def version_okay(v: str) -> bool:
            return self.prerelease or all(p.isdigit() for p in v.split("."))

        def sort_version(v: str) -> tuple:
            parts = []
            for part in v.split("."):
                if part.isdigit():
                    parts.append(int(part))
                else:
                    digit, rest = re.match(r"^(\d*)(.*)", part).groups()
                    if digit:
                        parts.append(int(digit))
                    parts.append(rest)
            return tuple(parts)

        installable_versions = {
            k for k, v in metadata["releases"].items() if version_okay(k) and not v[0].get("yanked")
        }
        releases = sorted(installable_versions, key=sort_version, reverse=True)

        return releases[0]

    def _decide_path(self) -> Path:
        if self.location is not None:
            return Path(self.location).expanduser().resolve()

        if MACOS:
            path = os.path.expanduser("~/Library/trusted-ai")
        else:
            path = os.getenv("XDG_DATA_HOME", os.path.expanduser("~/.local/share"))
            path = os.path.join(path, "trusted-ai")

        return Path(path)

    def _make_env(self) -> Path:
        venv_path = self._path / "venv"

        _echo(
            "Installing {} ({}): {}".format(
                colored("green", "trusted-ai", bold=True),
                colored("yellow", self.version),
                colored("cyan", "Creating virtual environment"),
            )
        )

        try:
            import venv

            venv.create(venv_path, clear=False, with_pip=True)
        except (ModuleNotFoundError, subprocess.CalledProcessError):
            try:
                import virtualenv
            except ModuleNotFoundError:
                python_version = f"{sys.version_info.major}.{sys.version_info.minor}"
                url = f"https://bootstrap.pypa.io/virtualenv/{python_version}/virtualenv.pyz"
                with TemporaryDirectory(prefix="trusted-ai-installer-") as tempdir:
                    virtualenv_zip = Path(tempdir) / "virtualenv.pyz"
                    urllib.request.urlretrieve(url, virtualenv_zip)
                    _call_subprocess([sys.executable, str(virtualenv_zip), str(venv_path)])
            else:
                virtualenv.cli_run([str(venv_path)])

        return venv_path

    def _install(self, venv_path: Path) -> None:
        _echo(
            "Installing {} ({}): {}".format(
                colored("green", "trusted-ai", bold=True),
                colored("yellow", self.version),
                colored("cyan", "Installing trusted-ai and dependencies"),
            )
        )

        # if WINDOWS:
        #     venv_python = venv_path / "Scripts/python.exe"
        # else:
        venv_python = venv_path / "bin" / "python"

        # Re-install the venv pip to ensure it's not DEBUNDLED
        # See issue/685
        try:
            _call_subprocess([str(venv_python), "-m", "ensurepip"])
        except SystemExit:
            pass
        _call_subprocess([str(venv_python), "-m", "pip", "install", "-IU", "pip"])

        if self.version:
            req = f"trusted-ai=={self.version}"
        else:
            req = "trusted-ai"
        args = [req] + [d for d in self.additional_deps if d]
        pip_cmd = [str(venv_python), "-Im", "pip", "install", *args]

        _call_subprocess(pip_cmd)

    def _make_bin(self, venv_path: Path) -> Path:
        if self.location:
            bin_path = self._path / "bin"
        else:
            userbase = Path(site.getuserbase())
            bin_path = userbase / "bin"

        _echo(
            "Installing {} ({}): {} {}".format(
                colored("green", "trusted-ai", bold=True),
                colored("yellow", self.version),
                colored("cyan", "Making binary at"),
                colored("green", str(bin_path)),
            )
        )
        bin_path.mkdir(parents=True, exist_ok=True)
        script = bin_path / "tai"
        target = venv_path / "bin" / "tai"

        if script.exists():
            script.unlink()
        try:
            script.symlink_to(target)
        except OSError:
            shutil.copy(target, script)
        return bin_path

    def _post_install(self, venv_path: Path, bin_path: Path) -> None:
        script = bin_path / "tai"
        subprocess.check_call([str(script), "--help"])
        print()
        _echo(
            "Successfully installed: {} ({}) at {}".format(
                colored("green", "trusted-ai", bold=True),
                colored("yellow", self.version),
                colored("cyan", str(script)),
            )
        )
        if not self.skip_add_to_path:
            _add_to_path(bin_path)
        self._write_output(venv_path, script)

    def _write_output(self, venv_path: Path, script: Path) -> None:
        if not self.output_path:
            return
        print("Writing output to", colored("green", self.output_path))
        output = {
            "trusted_ai_version": self.version,
            "trusted_ai_bin": str(script),
            "install_python_version": f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}",
            "install_location": str(venv_path),
        }
        with open(self.output_path, "w") as f:
            json.dump(output, f, indent=2)

    def install(self) -> None:
        venv = self._make_env()
        self._install(venv)
        bin_dir = self._make_bin(venv)
        self._post_install(venv, bin_dir)

    def uninstall(self) -> None:
        _echo(
            "Uninstalling {}: {}".format(
                colored("green", "trusted-ai", bold=True),
                colored("cyan", "Removing venv and script"),
            )
        )
        if self.location:
            bin_path = self._path / "bin"
        else:
            userbase = Path(site.getuserbase())
            bin_path = userbase / "bin"

        script = bin_path / "trusted-ai"

        shutil.rmtree(self._path / "venv")
        script.unlink()

        print()
        _echo("Successfully uninstalled")


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-v",
        "--version",
        help="Specify the version to be installed",
        default=os.getenv("TRUSTED_AI_VERSION"),
    )
    parser.add_argument(
        "--prerelease",
        action="store_true",
        help="Allow prereleases to be installed",
        default=os.getenv("TRUSTED_AI_PRERELEASE"),
    )
    parser.add_argument(
        "--remove",
        action="store_true",
        help="Remove the trusted-ai installation",
        default=os.getenv("TRUSTED_AI_REMOVE"),
    )
    parser.add_argument(
        "-p",
        "--path",
        help="Specify the location to install trusted-ai",
        default=os.getenv("TRUSTED_AI_HOME"),
    )
    parser.add_argument(
        "-d",
        "--dep",
        action="append",
        default=os.getenv("TRUSTED_AI_DEPS", "").split(","),
        help="Specify additional dependencies, can be given multiple times",
    )
    parser.add_argument(
        "--skip-add-to-path",
        action="store_true",
        help="Do not add binary to the PATH.",
        default=os.getenv("TRUSTED_AI_SKIP_ADD_TO_PATH"),
    )
    parser.add_argument("-o", "--output", help="Output file to write the installation info to")

    options = parser.parse_args()
    installer = Installer(
        location=options.path,
        version=options.version,
        prerelease=options.prerelease,
        additional_deps=options.dep,
        skip_add_to_path=options.skip_add_to_path,
        output_path=options.output,
    )
    if options.remove:
        installer.uninstall()
    else:
        installer.install()


if __name__ == "__main__":
    main()


import argparse
import shutil
from git import Repo
from pathlib import Path
import subprocess
import yaml

from tai.core.utils import color, create_template_user_config
from tai.core.gitea_controller import GiteaController


ADMIN_CONFIG_PATH = "/Users/Ekaterina/Documents/ISPwork/trusted-ai/tai/core/gitea_conf.yaml"
GITEA_URL = "git.ai.intra.ispras.ru/"
GITEA_CONFIG_PATH = "user_gitea_config.yaml"

def main():
    """Main function with running commands.

    :param mode: `new` for creation new package, `help` for text description, create-config for create new gitea config.
    :type mode: str

    :param pkg_name: name of your <package name> (for mode `new`)
    :type pkg_name: str

    :param --local-repo-path: path to folder where you want to create repo
    :type local-repo-path: str

    :param --config-path: path to config file with your creds
    :type config-path: str

    :param --username: gitea username
    :type username: str

    :param --email: gitea email
    :type email: str

    :param --full-name: gitea full name
    :type full-name: str

    :param --password: gitea password
    :type password: str
    """
    parser = argparse.ArgumentParser(prog="trusted-ai")
    parser.add_argument("mode", type=str, choices=["new", "help", "create-config"])
    parser.add_argument("pkg_name", type=str, nargs="?")
    parser.add_argument("--hand-setting", default=False, action='store_true')
    parser.add_argument("--local-repo-path", type=str, default=Path.cwd())
    parser.add_argument("--config-path", type=str, nargs="?", default=GITEA_CONFIG_PATH)

    parser.add_argument("--username", type=str, nargs="?", help="Username for Gitea.")
    parser.add_argument("--email", type=str, nargs="?", help="Email for Gitea.")
    parser.add_argument("--full-name", type=str, nargs="*", help="Full name for Gitea.")
    parser.add_argument("--password", type=str, nargs="?", help="Password for Gitea.")

    args = parser.parse_args()

    if args.mode == "help":
        print("Hello! \n It's TAI project, that will help you create your own little-package for MLM.")
        print(
            f"Type this command for creation new package:\n\t" + color.BOLD + color.GREEN + 
            f"tai new " + color.END + color.BLUE + color.BOLD +
            f"<package_name> --local-repo-path <your path> --config-path <your config>"
            + color.END)
        print("Add key --hand-setting, if you want to choose settings in pdm by yourself.")
        print(
            f"If you don't have configuration file for Gitea, yo can create it with command:\n\t" +
            color.BOLD + color.GREEN +
            f"tai create-config " + color.END + color.BLUE + color.BOLD +
            f"--config-path <path to config> " +
            f"--username <your username> --email <your email> --full-name <your full name> --password <your password>" +
            color.END
            )

    elif args.mode == "new":
        gc = GiteaController(admin_config_path=ADMIN_CONFIG_PATH)
        if args.config_path is None:
            conf_path = GITEA_CONFIG_PATH
        else:
            conf_path = args.config_path
        gc.create_user_config(user_config_path=conf_path)
        exist_repo = gc.check_users_repo(repo_name=args.pkg_name)
        if exist_repo==True:
            raise NameError(f"Repository with name {args.pkg_name} already exists!")
        else:
            gc.create_repo(repo_name=args.pkg_name)

        create_new_repo(package_name=args.pkg_name, local_repo_path=args.local_repo_path, gitea_controller=gc, hand_setting=args.hand_setting)

    elif args.mode == "create-config":
        if args.config_path is None:
            conf_path = GITEA_CONFIG_PATH
        else:
            conf_path = args.config_path
        full_name = " ".join(args.full_name) if args.full_name else args.full_name
        create_template_user_config(conf_path=conf_path,
                                    email=args.email,
                                    full_name=full_name,
                                    username=args.username,
                                    password=args.password,
                                    )
    else:
        raise ValueError("You can choose only these modes: new, help.")


def create_new_repo(package_name, local_repo_path, gitea_controller, hand_setting):
    """
    Create a new git repository with a package for your Model|Dataset|Executor.

    This function creates a repo in <local_repo_path> with the following structure::
        
        tai/
        ├── pyproject.toml  # pdm configuration file for your project
        ├── .gitignore
        ├── .venv           # virtual venv for your project
        └── contrib/
            └── <organization or username>/
                └── <package name>/
                    └── __init__.py

    :param package_name: Package name
    :type package_name: str

    :param local_repo_path: path to folder where you want to create repo
    :type local_repo_path: str

    :param gitea_controller: object of GiteaController
    :type gitea_controller: GiteaController
    """
    file_path = Path(local_repo_path) / "tai" / "contrib" / gitea_controller.user_config["username"] / package_name
    file_path.mkdir(exist_ok=True, parents=True)
    open(file_path / "__init__.py", "a").close()

    shutil.copyfile(Path(__file__).parent / ".gitignore_template", Path(local_repo_path) / ".gitignore")

    Repo.init(local_repo_path)
    repo = Repo(local_repo_path)
    repo.index.add(['.gitignore'])

    command = ["pdm", "init"]
    process = subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE, text=True, cwd=local_repo_path)

    if hand_setting is False:
        input_data_list = ["0", # TODO: add Python interpreter selection 
                        f"{gitea_controller.user_config['username']}_{package_name}",
                        "0.1.0",
                        "y",
                        "Some description...",
                        "0",
                        "MIT",
                        gitea_controller.user_config["username"],
                        gitea_controller.user_config["email"],
                        "==3.12.*"]

        for input_data in input_data_list:
            process.stdin.write(input_data + "\n")
            process.stdin.flush()

        output = process.communicate()
        print("Output:", "\n".join(output))

        process.stdin.close()
        process.stdout.close()
        process.stderr.close()

    else:
        subprocess.call(command, cwd=local_repo_path)

    shutil.rmtree(Path(local_repo_path) / "src")
    shutil.rmtree(Path(local_repo_path) / "tests")

    repo.index.add([".gitignore", "tai", "pyproject.toml"])

    username = gitea_controller.user_config['username']
    password = gitea_controller.user_config['password']

    repo_url = f"https://{username}:{password}@" + GITEA_URL + username + "/" + package_name + ".git"

    repo.config_writer().set_value("user", "name", gitea_controller.user_config["username"]).release()
    repo.config_writer().set_value("user", "email", gitea_controller.user_config["email"]).release()
    repo.index.commit("Initial commit.")
    if "origin" in repo.remotes:
        origin = repo.remotes.origin
    else:
        origin = repo.create_remote("origin", repo_url)
    repo.git.push("--set-upstream", origin, "master")


if __name__ == "__main__":
    main()


import giteapy
from giteapy.rest import ApiException
from tai.core.utils import read_config


class GiteaController:
    def __init__(self, admin_config_path: str):
        config = read_config(admin_config_path)
        gitea_configuration = create_conf(config["admin_token"], config["username"], config["password"])
        self.admin_api_object = giteapy.ApiClient(gitea_configuration)
        self.admin_api = giteapy.AdminApi(self.admin_api_object)
        self.user_api = None
        self.user_config = None
    
    def create_user_config(self, user_config_path: str):
        self.user_config = read_config(user_config_path)
        self.user_config_path = user_config_path
        configuration = giteapy.Configuration()
        configuration.host = "https://git.ai.intra.ispras.ru/api/v1"
        configuration.username = self.user_config["username"]
        configuration.password = self.user_config["password"]
        # if "token" in self.user_config:
        #     configuration.api_key =  self.user_config["token"]
        self.user_api = giteapy.UserApi(giteapy.ApiClient(configuration))


    def check_user_existing(self, username):
        user_exist = None
        try:
            users = self.admin_api.admin_get_all_users()
            for user in users:
                if user.login == username:
                    user_exist = True
                    break
            user_exist = False
        except giteapy.rest.ApiException as error:
            user_exist = False
            print(error.body)
        return user_exist
    
    def create_new_user(self):
        status = None
        try:
            if self.user_config is None:
                raise ValueError("Create user config! Call GiteaController.create_user_config(<config_path>).")
            user_creds = {
                "email": self.user_config["email"],
                "full_name": self.user_config["full_name"],
                "password": self.user_config["password"],
                "username": self.user_config["username"],
                "must_change_password": False,
            }
            self.admin_api.admin_create_user(body=user_creds)
            status = True
        except giteapy.rest.ApiException as error:
            status = False
            print(error.body)
        return status
    
    def create_repo(self, repo_name: str):
        status = None
        if self.user_config is None:
            raise ValueError("Create user config! Call GiteaController.create_user_config(<config_path>).")
        user_exist = self.check_user_existing(username=self.user_config["username"])
        if user_exist is False:
            print("Creating new user.")
            self.create_new_user()
        try:
            repository = giteapy.CreateRepoOption(name=repo_name)
            self.admin_api.admin_create_repo(username=self.user_config["username"], repository=repository)
            status = True
        except giteapy.rest.ApiException as error:
            status = False
            print(error)
        return status
    
    def remove_repo(self, repo_name):
        status = None
        try:
            api_instance = giteapy.RepositoryApi(self.admin_api_object)
            api_instance.repo_delete(owner=self.user_config["username"],
                                     repo=repo_name)
            status = True
        except ApiException as error:
            status = False
            print("Exception when calling AdminApi->repo_delete: %s\n" % error)
        return status
    
    def remove_user(self, username):
        repos = self.user_api.user_current_list_repos()
        for rep in repos:
            print(rep)
            self.remove_repo(rep.name)
        self.admin_api.admin_delete_user(username)
    
    def check_users_repo(self, repo_name):
        exist_status = None
        try:
            repos = self.user_api.user_current_list_repos()
            for repo in repos:
                if repo.name == repo_name:
                    exist_status = True
                    break
            exist_status = False
        except ApiException as error:
            print("Exception when calling UserApi->user_current_list_repos: %s\n" % error)
        return exist_status


def create_conf(api_key, username, password):
    configuration = giteapy.Configuration()
    configuration.host = "https://git.ai.intra.ispras.ru/api/v1"
    configuration.username = username
    configuration.password = password
    configuration.api_key['access_token'] = api_key
    configuration.api_key['sudo'] = api_key
    configuration.api_key['Authorization'] = api_key
    configuration.api_key['Sudo'] = api_key
    configuration.api_key['token'] = api_key

    return configuration

if __name__ == "__main__":
    # TODO: change getting admin config
    gc = GiteaController(admin_config_path="/Users/Ekaterina/Documents/ISPwork/trusted-ai/tai/core/gitea_conf.yaml")
    # gc.check_user_existing(username="pp")
    gc.create_user_config("/Users/Ekaterina/Documents/ISPwork/trusted-ai/user_gitea_config.yaml")

    # status = gc.create_repo(repo_name="other_repo")
    # # print(status)

    # gc.check_users_repo(repo_name="other_repo")

    # gc.remove_repo("kekrepo")

    # gc.admin_api.admin_delete_user("test_user_6")

    gc.remove_user("test_user")